"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.convertBoard = exports.convertBoardToArray = exports.convertShape = void 0;
var spectra_1 = require("./spectra");
var svg_arc_1 = require("./svg-arc");
function getLayerName(id, conversionState) {
    var layers = {
        1: 'F.Cu',
        2: 'B.Cu',
        3: 'F.SilkS',
        4: 'B.SilkS',
        5: 'F.Paste',
        6: 'B.Paste',
        7: 'F.Mask',
        8: 'B.Mask',
        10: 'Edge.Cuts',
        11: 'Edge.Cuts',
        12: 'Cmts.User',
        13: 'F.Fab',
        14: 'B.Fab',
        15: 'Dwgs.User',
    };
    if (id in layers) {
        return layers[id];
    }
    // Inner layers: 21 -> In1.Cu
    var intId = parseInt(id, 10);
    if (intId >= 21 && intId <= 50) {
        var innerLayerId = intId - 20;
        conversionState.innerLayers = Math.max(conversionState.innerLayers, innerLayerId);
        return "In" + innerLayerId + ".Cu";
    }
    if (intId >= 99 && intId < 200) {
        console.warn("Warning: unsupported layer id: " + intId);
        return null;
    }
    throw new Error("Missing layer id: " + id);
}
function kiUnits(value) {
    if (typeof value === 'string') {
        value = parseFloat(value);
    }
    return value * 10 * 0.0254;
}
function kiAngle(value, parentAngle) {
    if (value) {
        var angle = parseFloat(value) + (parentAngle || 0);
        if (!isNaN(angle)) {
            return angle > 180 ? -(360 - angle) : angle;
        }
    }
    return null;
}
function rotate(_a, degrees) {
    var x = _a.x, y = _a.y;
    var radians = (degrees / 180) * Math.PI;
    return {
        x: x * Math.cos(radians) - y * Math.sin(radians),
        y: x * Math.sin(radians) + y * Math.cos(radians),
    };
}
function kiCoords(x, y, transform) {
    if (transform === void 0) { transform = { x: 0, y: 0, angle: 0 }; }
    return rotate({
        x: kiUnits(parseFloat(x) - 4000) - transform.x,
        y: kiUnits(parseFloat(y) - 3000) - transform.y,
    }, transform.angle || 0);
}
function kiAt(x, y, angle, transform) {
    var coords = kiCoords(x, y, transform);
    return ['at', coords.x, coords.y, kiAngle(angle)];
}
function kiStartEnd(startX, startY, endX, endY, parentCoords) {
    var start = kiCoords(startX, startY, parentCoords);
    var end = kiCoords(endX, endY, parentCoords);
    return [
        ['start', start.x, start.y],
        ['end', end.x, end.y],
    ];
}
function isCopper(layerName) {
    return layerName.endsWith('.Cu');
}
function getNetId(_a, netName) {
    var nets = _a.nets;
    if (!netName) {
        return -1;
    }
    var index = nets.indexOf(netName);
    if (index >= 0) {
        return index;
    }
    nets.push(netName);
    return nets.length - 1;
}
function convertVia(args, conversionState, parentCoords) {
    var x = args[0], y = args[1], diameter = args[2], net = args[3], drill = args[4], id = args[5], locked = args[6];
    return [
        'via',
        kiAt(x, y, undefined, parentCoords),
        ['size', kiUnits(diameter)],
        ['drill', kiUnits(drill) * 2],
        ['layers', 'F.Cu', 'B.Cu'],
        ['net', getNetId(conversionState, net)],
    ];
}
function convertPadToVia(args, conversionState, parentCoords) {
    var shape = args[0], x = args[1], y = args[2], holeRadius = args[3], height = args[4], layerId = args[5], net = args[6], num = args[7], drill = args[8], points = args[9], rotation = args[10], id = args[11], holeLength = args[12], holePoints = args[13], plated = args[14], locked = args[15];
    var size = kiUnits(holeRadius);
    var drillHoleLength = holeLength === '0' ? null : kiUnits(holeLength);
    if (shape !== 'ELLIPSE') {
        return [
            'module',
            'AutoGenerated:Pad_' + size.toFixed(2) + 'mm',
            locked === '1' ? 'locked' : null,
            ['layer', 'F.Cu'],
            kiAt(x, y),
            ['attr', 'virtual'],
            ['fp_text', 'reference', '', ['at', 0, 0], ['layer', 'F.SilkS']],
            ['fp_text', 'value', '', ['at', 0, 0], ['layer', 'F.SilkS']],
            convertPad(args, conversionState, __assign(__assign({}, kiCoords(x, y)), { angle: 0 })),
        ];
    }
    return [
        'via',
        kiAt(x, y, undefined, parentCoords),
        ['size', kiUnits(holeRadius)],
        ['drill', kiUnits(drill) * 2],
        ['layers', 'F.Cu', 'B.Cu'],
        ['net', getNetId(conversionState, net)],
    ];
}
function convertTrack(args, conversionState, objName, parentCoords) {
    if (objName === void 0) { objName = 'segment'; }
    var width = args[0], layer = args[1], net = args[2], coords = args[3], id = args[4], locked = args[5];
    var netId = getNetId(conversionState, net);
    var coordList = coords.split(' ');
    var result = [];
    var layerName = getLayerName(layer, conversionState);
    if (!layerName) {
        return [];
    }
    var lineType = objName === 'segment' && !isCopper(layerName) ? 'gr_line' : objName;
    for (var i = 0; i < coordList.length - 2; i += 2) {
        result.push(__spreadArrays([
            lineType
        ], kiStartEnd(coordList[i], coordList[i + 1], coordList[i + 2], coordList[i + 3], parentCoords), [
            ['width', kiUnits(width)],
            ['layer', layerName],
            isCopper(layerName) && netId > 0 ? ['net', netId] : null,
            locked === '1' ? ['status', 40000] : null,
        ]));
    }
    return result;
}
function textLayer(layer, conversionState, footprint, isName) {
    var layerName = getLayerName(layer, conversionState);
    if (layerName && footprint && isName) {
        return layerName.replace('.SilkS', '.Fab');
    }
    else {
        return layerName;
    }
}
function convertText(args, conversionState, objName, parentCoords) {
    if (objName === void 0) { objName = 'gr_text'; }
    var type = args[0], // N/P/L (Name/Prefix/Label)
    x = args[1], y = args[2], lineWidth = args[3], angle = args[4], mirror = args[5], layer = args[6], net = args[7], fontSize = args[8], text = args[9], path = args[10], display = args[11], id = args[12], font = args[13], locked = args[14];
    var layerName = textLayer(layer, conversionState, objName === 'fp_text', type === 'N');
    if (!layerName) {
        return null;
    }
    var fontTable = {
        'NotoSerifCJKsc-Medium': { width: 0.8, thickness: 0.3 },
        'NotoSansCJKjp-DemiLight': { width: 0.6, thickness: 0.5 },
    };
    var fontMultiplier = font in fontTable ? fontTable[font] : { width: 1, thickness: 1 };
    var actualFontSize = kiUnits(fontSize) * fontMultiplier.width;
    return [
        objName,
        objName === 'fp_text' ? (type === 'P' ? 'reference' : 'value') : null,
        text,
        kiAt(x, y, angle, parentCoords),
        ['layer', layerName],
        display === 'none' ? 'hide' : null,
        [
            'effects',
            [
                'font',
                ['size', actualFontSize, actualFontSize],
                ['thickness', kiUnits(lineWidth) * fontMultiplier.thickness],
            ],
            ['justify', 'left', layerName[0] === 'B' ? 'mirror' : null],
        ],
    ];
}
function convertArc(args, conversionState, objName, transform) {
    if (objName === void 0) { objName = 'gr_arc'; }
    var width = args[0], layer = args[1], net = args[2], path = args[3], _ = args[4], id = args[5], locked = args[6];
    var layerName = getLayerName(layer, conversionState);
    if (!layerName) {
        return null;
    }
    var pathMatch = /^M\s*([-\d.\s]+)A\s*([-\d.\s]+)$/.exec(path.replace(/[,\s]+/g, ' '));
    if (!pathMatch) {
        console.warn("Invalid arc path: " + path);
        return null;
    }
    var match = pathMatch[0], startPoint = pathMatch[1], arcParams = pathMatch[2];
    var _a = startPoint.split(' '), startX = _a[0], startY = _a[1];
    var _b = arcParams.split(' '), svgRx = _b[0], svgRy = _b[1], xAxisRotation = _b[2], largeArc = _b[3], sweep = _b[4], endX = _b[5], endY = _b[6];
    var start = kiCoords(startX, startY, transform);
    var end = kiCoords(endX, endY, transform);
    var _c = rotate({ x: kiUnits(svgRx), y: kiUnits(svgRy) }, (transform === null || transform === void 0 ? void 0 : transform.angle) || 0), rx = _c.x, ry = _c.y;
    var _d = svg_arc_1.computeArc(start.x, start.y, rx, ry, parseFloat(xAxisRotation), largeArc === '1', sweep === '1', end.x, end.y), cx = _d.cx, cy = _d.cy, extent = _d.extent;
    var endPoint = sweep === '1' ? start : end;
    if (isNaN(cx) || isNaN(cy)) {
        console.warn("Invalid arc: " + path);
        return null;
    }
    return [
        objName,
        ['start', cx, cy],
        ['end', endPoint.x, endPoint.y],
        ['angle', Math.abs(extent)],
        ['width', kiUnits(width)],
        ['layer', layerName],
    ];
}
function getDrill(holeRadius, holeLength) {
    if (holeRadius && holeLength) {
        return ['drill', 'oval', holeRadius * 2, holeLength];
    }
    if (holeRadius) {
        return ['drill', holeRadius * 2];
    }
    return null;
}
function isRectangle(points) {
    if (points.length !== 8) {
        return false;
    }
    var eq = function (a, b) { return Math.abs(a - b) < 0.01; };
    var x1 = points[0], y1 = points[1], x2 = points[2], y2 = points[3], x3 = points[4], y3 = points[5], x4 = points[6], y4 = points[7];
    return ((eq(x1, x2) && eq(y2, y3) && eq(x3, x4) && eq(y4, y1)) ||
        (eq(y1, y2) && eq(x2, x3) && eq(y3, y4) && eq(x4, x1)));
}
function rectangleSize(points, rotation) {
    var x1 = points[0], y1 = points[1], x2 = points[2], y2 = points[3], x3 = points[4], y3 = points[5], x4 = points[6], y4 = points[7];
    var width = Math.max(x1, x2, x3, x4) - Math.min(x1, x2, x3, x4);
    var height = Math.max(y1, y2, y3, y4) - Math.min(y1, y2, y3, y4);
    return Math.round(Math.abs(rotation)) % 180 === 90 ? [height, width] : [width, height];
}
function convertPad(args, conversionState, transform) {
    var shape = args[0], x = args[1], y = args[2], width = args[3], height = args[4], layerId = args[5], net = args[6], num = args[7], holeRadius = args[8], points = args[9], rotation = args[10], id = args[11], holeLength = args[12], holePoints = args[13], plated = args[14], locked = args[15];
    var padShapes = {
        ELLIPSE: 'circle',
        RECT: 'rect',
        OVAL: 'oval',
        POLYGON: 'custom',
    };
    var centerCoords = kiCoords(x, y);
    var polygonTransform = {
        x: centerCoords.x,
        y: centerCoords.y,
        angle: parseFloat(rotation),
    };
    var pointList = points.split(' ').map(parseFloat);
    var pointsAreRectangle = padShapes[shape] === 'custom' && isRectangle(pointList);
    var actualShape = pointsAreRectangle ? 'RECT' : shape;
    var isCustomShape = padShapes[actualShape] === 'custom';
    if (isCustomShape && !points.length) {
        console.warn("PAD " + id + " is a polygon, but has no points defined");
        return null;
    }
    var netId = getNetId(conversionState, net);
    var layers = {
        1: ['F.Cu', 'F.Paste', 'F.Mask'],
        2: ['B.Cu', 'B.Paste', 'B.Mask'],
        11: ['*.Cu', '*.Paste', '*.Mask'],
    };
    var _a = pointsAreRectangle
        ? rectangleSize(pointList, parseFloat(rotation))
        : [width, height], actualWidth = _a[0], actualHeight = _a[1];
    var padNum = parseInt(num, 10);
    return [
        'pad',
        isNaN(padNum) ? num : padNum,
        kiUnits(holeRadius) > 0 ? 'thru_hole' : 'smd',
        padShapes[actualShape],
        kiAt(x, y, rotation, transform),
        ['size', Math.max(kiUnits(actualWidth), 0.01), Math.max(kiUnits(actualHeight), 0.01)],
        __spreadArrays(['layers'], layers[layerId]),
        getDrill(kiUnits(holeRadius), kiUnits(holeLength)),
        netId > 0 ? ['net', netId, net] : null,
        isCustomShape
            ? [
                'primitives',
                [
                    'gr_poly',
                    __spreadArrays(['pts'], pointListToPolygon(points.split(' '), polygonTransform)),
                    ['width', 0.1],
                ],
            ]
            : null,
    ];
}
function convertLibHole(args, transform) {
    var x = args[0], y = args[1], radius = args[2], id = args[3], locked = args[4];
    var size = kiUnits(radius) * 2;
    return [
        'pad',
        '',
        'np_thru_hole',
        'circle',
        kiAt(x, y, undefined, transform),
        ['size', size, size],
        ['drill', size],
        ['layers', '*.Cu', '*.Mask'],
    ];
}
function convertCircle(args, conversionState, type, parentCoords) {
    if (type === void 0) { type = 'gr_circle'; }
    var x = args[0], y = args[1], radius = args[2], strokeWidth = args[3], layer = args[4], id = args[5], locked = args[6];
    var layerName = getLayerName(layer, conversionState);
    if (!layerName) {
        return null;
    }
    var center = kiCoords(x, y, parentCoords);
    return [
        type,
        ['center', center.x, center.y],
        ['end', center.x + kiUnits(radius), center.y],
        ['layer', layerName],
        ['width', kiUnits(strokeWidth)],
    ];
}
function pointListToPolygon(points, parentCoords) {
    var polygonPoints = [];
    for (var i = 0; i < points.length; i += 2) {
        var coords = kiCoords(points[i], points[i + 1], parentCoords);
        polygonPoints.push(['xy', coords.x, coords.y]);
    }
    return polygonPoints;
}
function pathToPolygon(path, parentCoords) {
    if (path.indexOf('A') >= 0) {
        console.warn('Warning: SOLIDREGION with arcs/circles are not supported yet!');
        return null;
    }
    var points = path.split(/[ ,LM]/).filter(function (p) { return !isNaN(parseFloat(p)); });
    return pointListToPolygon(points, parentCoords);
}
function convertPolygon(args, conversionState, parentCoords) {
    var layerId = args[0], net = args[1], path = args[2], type = args[3], id = args[4], locked = args[7];
    if (type !== 'solid') {
        console.warn("Warning: unsupported SOLIDREGION type in footprint: " + type);
        return null;
    }
    var layerName = getLayerName(layerId, conversionState);
    if (!layerName) {
        return null;
    }
    var polygonPoints = pathToPolygon(path, parentCoords);
    if (!polygonPoints) {
        return null;
    }
    return ['fp_poly', __spreadArrays(['pts'], polygonPoints), ['layer', layerName], ['width', 0]];
}
function convertLib(args, conversionState) {
    var x = args[0], y = args[1], attributes = args[2], rotation = args[3], importFlag = args[4], id = args[5], locked = args[9];
    var shapeList = args.join('~').split('#@$').slice(1);
    var attrList = attributes.split('`');
    var attrs = {};
    for (var i = 0; i < attrList.length; i += 2) {
        attrs[attrList[i]] = attrList[i + 1];
    }
    var shapes = [];
    var transform = __assign(__assign({}, kiCoords(x, y)), { angle: kiAngle(rotation) });
    for (var _i = 0, shapeList_1 = shapeList; _i < shapeList_1.length; _i++) {
        var shape = shapeList_1[_i];
        var _a = shape.split('~'), type = _a[0], shapeArgs = _a.slice(1);
        if (type === 'TRACK') {
            shapes.push.apply(shapes, convertTrack(shapeArgs, conversionState, 'fp_line', transform));
        }
        else if (type === 'TEXT') {
            shapes.push(convertText(shapeArgs, conversionState, 'fp_text', transform));
        }
        else if (type === 'ARC') {
            shapes.push(convertArc(shapeArgs, conversionState, 'fp_arc', transform));
        }
        else if (type === 'HOLE') {
            shapes.push(convertLibHole(shapeArgs, transform));
        }
        else if (type === 'PAD') {
            shapes.push(convertPad(shapeArgs, conversionState, transform));
        }
        else if (type === 'CIRCLE') {
            shapes.push(convertCircle(shapeArgs, conversionState, 'fp_circle', transform));
        }
        else if (type === 'SOLIDREGION') {
            shapes.push(convertPolygon(shapeArgs, conversionState, transform));
        }
        else {
            console.warn("Warning: unsupported shape " + type + " in footprint " + id);
        }
    }
    shapes.push([
        'fp_text',
        'user',
        id,
        ['at', 0, 0],
        ['layer', 'Cmts.User'],
        ['effects', ['font', ['size', 1, 1], ['thickness', 0.15]]],
    ]);
    var modAttrs = [];
    var isSmd = shapes.some(function (shape) { return shape && shape[0] === 'pad' && shape[2] === 'smd'; });
    if (isSmd) {
        modAttrs.push(['attr', 'smd']);
    }
    var footprintName = "easyeda:" + (attrs.package || id);
    return __spreadArrays([
        'module',
        footprintName,
        locked === '1' ? 'locked' : null,
        ['layer', 'F.Cu'],
        kiAt(x, y, rotation)
    ], modAttrs, shapes);
}
function convertCopperArea(args, conversionState) {
    var strokeWidth = args[0], layerId = args[1], net = args[2], path = args[3], clearanceWidth = args[4], fillStyle = args[5], id = args[6], thermal = args[7], keepIsland = args[8], copperZone = args[9], locked = args[10];
    var netId = getNetId(conversionState, net);
    var layerName = getLayerName(layerId, conversionState);
    if (!layerName) {
        return null;
    }
    // fill style: solid/none
    // id: gge27
    // thermal: spoke/direct
    var pointList = path.split(/[ ,LM]/).filter(function (p) { return !isNaN(parseFloat(p)); });
    var polygonPoints = [];
    for (var i = 0; i < pointList.length; i += 2) {
        var coords = kiCoords(pointList[i], pointList[i + 1]);
        polygonPoints.push(['xy', coords.x, coords.y]);
    }
    return [
        'zone',
        ['net', netId],
        ['net_name', net],
        ['layer', layerName],
        ['hatch', 'edge', 0.508],
        ['connect_pads', ['clearance', kiUnits(clearanceWidth)]],
        // TODO (min_thickness 0.254)
        // TODO (fill yes (arc_segments 32) (thermal_gap 0.508) (thermal_bridge_width 0.508))
        ['polygon', __spreadArrays(['pts'], polygonPoints)],
    ];
}
function convertSolidRegion(args, conversionState) {
    var layerId = args[0], net = args[1], path = args[2], type = args[3], id = args[4], locked = args[5];
    var layerName = getLayerName(layerId, conversionState);
    if (!layerName) {
        return null;
    }
    var polygonPoints = pathToPolygon(path);
    var netId = getNetId(conversionState, net);
    if (!polygonPoints) {
        return null;
    }
    switch (type) {
        case 'cutout':
            return [
                'zone',
                ['net', netId],
                ['net_name', ''],
                ['hatch', 'edge', 0.508],
                ['layer', layerName],
                ['keepout', ['tracks', 'allowed'], ['vias', 'allowed'], ['copperpour', 'not_allowed']],
                ['polygon', __spreadArrays(['pts'], polygonPoints)],
            ];
        case 'solid':
        case 'npth':
            return [
                'gr_poly',
                __spreadArrays(['pts'], polygonPoints),
                ['layer', layerName],
                ['width', 0],
            ];
        default:
            console.warn("Warning: unsupported SOLIDREGION type " + type);
            return null;
    }
}
function convertHole(args) {
    var x = args[0], y = args[1], radius = args[2], id = args[3], locked = args[4];
    var size = kiUnits(radius) * 2;
    return [
        'module',
        "AutoGenerated:MountingHole_" + size.toFixed(2) + "mm",
        locked === '1' ? 'locked' : null,
        ['layer', 'F.Cu'],
        kiAt(x, y),
        ['attr', 'virtual'],
        ['fp_text', 'reference', '', ['at', 0, 0], ['layer', 'F.SilkS']],
        ['fp_text', 'value', '', ['at', 0, 0], ['layer', 'F.SilkS']],
        [
            'pad',
            '',
            'np_thru_hole',
            'circle',
            ['at', 0, 0],
            ['size', size, size],
            ['drill', size],
            ['layers', '*.Cu', '*.Mask'],
        ],
    ];
}
function convertShape(shape, conversionState) {
    var _a = shape.split('~'), type = _a[0], args = _a.slice(1);
    switch (type) {
        case 'VIA':
            return [convertVia(args, conversionState)];
        case 'TRACK':
            return convertTrack(args, conversionState);
        case 'TEXT':
            return [convertText(args, conversionState)];
        case 'ARC':
            return [convertArc(args, conversionState)];
        case 'COPPERAREA':
            return [convertCopperArea(args, conversionState)];
        case 'SOLIDREGION':
            return [convertSolidRegion(args, conversionState)];
        case 'CIRCLE':
            return [convertCircle(args, conversionState)];
        case 'HOLE':
            return [convertHole(args)];
        case 'LIB':
            return [convertLib(args, conversionState)];
        case 'PAD':
            return [convertPadToVia(args, conversionState)];
        default:
            console.warn("Warning: unsupported shape " + type);
            return null;
    }
}
exports.convertShape = convertShape;
function flatten(arr) {
    var _a;
    return (_a = []).concat.apply(_a, arr);
}
function convertBoardToArray(input) {
    var nets = (input.routerRule || { nets: [] }).nets;
    var conversionState = { nets: nets, innerLayers: 0 };
    nets.unshift(''); // Kicad expects net 0 to be empty
    var shapes = flatten(input.shape.map(function (shape) { return convertShape(shape, conversionState); }));
    var outputObjs = __spreadArrays(nets.map(function (net, idx) { return ['net', idx, net]; }), shapes).filter(function (obj) { return obj != null; });
    var innerLayers = [];
    for (var i = 1; i <= conversionState.innerLayers; i++) {
        innerLayers.push([i, "In" + i + ".Cu", 'signal']);
    }
    var layers = __spreadArrays([
        [0, 'F.Cu', 'signal']
    ], innerLayers, [
        [31, 'B.Cu', 'signal'],
        [32, 'B.Adhes', 'user'],
        [33, 'F.Adhes', 'user'],
        [34, 'B.Paste', 'user'],
        [35, 'F.Paste', 'user'],
        [36, 'B.SilkS', 'user'],
        [37, 'F.SilkS', 'user'],
        [38, 'B.Mask', 'user'],
        [39, 'F.Mask', 'user'],
        [40, 'Dwgs.User', 'user'],
        [41, 'Cmts.User', 'user'],
        [42, 'Eco1.User', 'user'],
        [43, 'Eco2.User', 'user'],
        [44, 'Edge.Cuts', 'user'],
        [45, 'Margin', 'user'],
        [46, 'B.CrtYd', 'user'],
        [47, 'F.CrtYd', 'user'],
        [48, 'B.Fab', 'user', 'hide'],
        [49, 'F.Fab', 'user', 'hide'],
    ]);
    return __spreadArrays([
        'kicad_pcb',
        ['version', 20171130],
        ['host', 'pcbnew', '(5.1.5)-3'],
        ['page', 'A4'],
        __spreadArrays(['layers'], layers)
    ], outputObjs);
}
exports.convertBoardToArray = convertBoardToArray;
function convertBoard(board) {
    return spectra_1.encodeObject(convertBoardToArray(board));
}
exports.convertBoard = convertBoard;
//# sourceMappingURL=board.js.map